import os
import collections
import datetime
import click

from morenines.util import get_hash


class Index(object):
    reader_version = 1

    @classmethod
    def _check_version(cls, headers):
        try:
            version = headers['version']
        except KeyError as e:
            missing_key = e.args[0]
            raise Exception("Invalid file format: missing required header '{}'".format(missing_key))

        if version != str(cls.reader_version):
            raise Exception("Unsupported file format version: file is {}, parser is {}".format(version, cls.reader_version))

    def __init__(self, path_prefix):
        self.path_prefix = path_prefix
        self.files = collections.OrderedDict()

    def read(self, path):
        with click.open_file(path, 'r') as stream:
            headers = parse_headers(stream)

            self._check_version(headers)

            self.files = parse_files(stream)

    def add(self, paths):
        for path in paths:
            # To hash the file, we need its absolute path
            abs_path = os.path.join(self.path_prefix, path)

            # We store the relative path in the index, not the absolute
            self.files[path] = get_hash(abs_path)

    def remove(self, paths):
        for path in paths:
            del self.files[path]

    def write(self, stream):
        headers = [
            ('version', self.reader_version),
            ('date', datetime.datetime.utcnow().isoformat()),
        ]

        for key, value in headers:
            stream.write("{}: {}\n".format(key, value))

        # Separate the headers from the files list with a blank line
        stream.write("\n")

        # Write files and hashes -- but sort the paths before writing
        for path in sorted(self.files):
            stream.write("{} {}\n".format(self.files[path], path))


##############################
# Index file parsing functions
##############################

def split_lines(lines, delim, num_fields):
    """Split each element in the sequence 'lines' into its component fields."""
    for line in lines:
        if line == '\n':
            return

        yield [field.strip() for field in line.split(delim, num_fields - 1)]

def parse_headers(file_):
    """Parse header lines of the form 'key: value'"""
    return {key: value for key, value in split_lines(file_, ':', 2)}

def parse_files(file_):
    """Parse file lines of the form 'HASHVALUE /path/to/file"""
    return {path:hash_ for hash_, path in split_lines(file_, ' ', 2)}



import os
import sys
import hashlib

def abort():
    """Stop exectution and return nonzero error code"""
    sys.exit(1)


def get_files(root_path, ignores, save_ignored_paths=False):
    paths = []
    ignored = []

    for dirpath, dirnames, filenames in os.walk(root_path):
        ignored_dirs = [d for d in dirnames if ignores.match(d)]

        # Only walk ignored dirs if we're saving all ignored paths
        if save_ignored_paths:
            # Directories end in a slash
            ignored.extend([path + '/' for path in rel_paths_iter(ignored_dirs, dirpath, root_path)])

        # Prune ignored subdirs of current dir in-place
        dirnames[:] = [d for d in dirnames if d not in ignored_dirs]

        for path in rel_paths_iter(filenames, dirpath, root_path):
            if ignores.match(path):
                if save_ignored_paths:
                    ignored.append(path)
                continue
            else:
                paths.append(path)

    return paths, ignored


def rel_paths_iter(names, parent_dir_path, root_path):
    for name in names:
        # We want the full path of the file/dir, not its name
        path = os.path.join(parent_dir_path, name)

        # That path must be relative to the root, not absolute
        path = os.path.relpath(path, root_path)

        yield path


def get_hash(path):
    h = hashlib.sha1()

    with open(path, 'rb') as f:
        h.update(f.read())

    return h.hexdigest()


def get_new_and_missing(repo, include_ignored=False):
    current_files, ignored_files = get_files(repo.path, repo.ignore, include_ignored)

    new_files = [path for path in current_files if path not in repo.index.files]

    missing_files = [path for path in repo.index.files.iterkeys() if path not in current_files]

    return new_files, missing_files, ignored_files



import click
import os
import sys

from morenines.index import Index
from morenines.ignores import Ignores
from morenines.repository import Repository
from morenines.util import get_files, get_hash, get_new_and_missing, abort
from morenines.output import info, success, warning, error, print_filelists


pass_repository = click.make_pass_decorator(Repository, ensure=True)

DEFAULT_PATH = os.getcwd()

def repo_path_callback(ctx, param, value):
    repo = ctx.ensure_object(Repository)
    if value:
        repo.open(value)
    else:
        repo.open(DEFAULT_PATH)
    return value


_common_params = {
    'new_repo_path' : click.argument("repo_path", required=False, default=DEFAULT_PATH, type=click.Path(resolve_path=True)),
    'repo_path': click.argument("repo_path", expose_value=False, required=False, callback=repo_path_callback, type=click.Path(resolve_path=True)),
    'ignored': click.option('-i', '--ignored/--no-ignored', 'show_ignored', default=False, help="Enable/disable showing files ignored by the ignores patterns."),
    'color': click.option('--color/--no-color', 'show_color', default=True, help="Enable/disable colorized output."),
}

def common_params(*param_names):
    def real_decorator(func):
        for param_name in param_names:
            func = _common_params[param_name](func)
        return func

    return real_decorator


@click.group()
def main():
    """A tool to track whether the content of files has changed."""
    pass


@main.command(short_help="Initialize a new morenines repository")
@common_params('new_repo_path')
@click.pass_context
def init(ctx, repo_path):
    repo = Repository()

    repo.create(repo_path)

    ctx.obj = repo

    success("Initialized empty morenines repository in {}".format(repo.mn_dir_path))


@main.command(short_help="Write a new index file")
@common_params('repo_path')
@pass_repository
def create(repo):
    """Write a new index file with the hashes of files under it."""

    if os.path.isfile(repo.index_path):
        error("Index file already exists: {}".format(repo.index_path))
        error("(To update an existing index, use the 'update' command)")
        abort()

    files, ignored = get_files(repo.path, repo.ignore)

    repo.index.add(files)

    with click.open_file(repo.index_path, mode='w') as stream:
        repo.index.write(stream)

    success('Wrote index file {}'.format(repo.index_path))


@main.command(short_help="Update an existing index file")
@common_params('repo_path')
@pass_repository
@click.option('--add-new/--no-add-new', default=False, help="Hash and add any files that aren't in the index")
@click.option('--remove-missing/--no-remove-missing', default=False, help="Delete any the hashes of any files in the index that no longer exist.")
def update(repo, add_new, remove_missing):
    """Update an existing index file with new file hashes, missing files removed, etc."""
    new_files, missing_files, ignored_files = get_new_and_missing(repo)

    if add_new:
        repo.index.add(new_files)

    if remove_missing is True:
        repo.index.remove(missing_files)

    if not any([new_files, missing_files]):
        info("Index is up-to-date (no new or missing files)")
    elif add_new or remove_missing:
        with click.open_file(repo.index_path, mode='w') as stream:
            repo.index.write(stream)
        success("Wrote index file {}".format(repo.index_path))
    else:
        warning("No action taken (use '--add-new' or '--remove-missing' to change the index)")


@main.command(short_help="Show new, missing or ignored files")
@common_params('repo_path', 'ignored', 'color')
@click.option('--verify/--no-verify', default=False, help="Re-hash all files in index and check for changes")
@pass_repository
@click.pass_context
def status(ctx, repo, show_ignored, show_color, verify):
    """Show any new files not in the index, index files that are missing, or ignored files."""
    new_files, missing_files, ignored_files = get_new_and_missing(repo, show_ignored)

    changed_files = []

    if verify:
        for path, old_hash in repo.index.files.iteritems():
            if path in missing_files:
                continue

            current_hash = get_hash(os.path.join(repo.path, path))

            if current_hash != old_hash:
                changed_files.append(path)

    ctx.color = show_color

    print_filelists(new_files, changed_files, missing_files, ignored_files)


@main.command(name='edit-ignores', short_help="Open the ignores file in an editor")
@common_params('repo_path')
@pass_repository
def edit_ignores(repo):
    """Open an existing or a new ignores file in an editor."""

    click.edit(filename=repo.ignore_path)


if __name__ == '__main__':
    main()



import os

from morenines import output
from morenines import util

from morenines.index import Index
from morenines.ignores import Ignores


NAMES = {
    'mn_dir': '.morenines',
    'index': 'index',
    'ignore': 'ignore',
}

DEFAULT_IGNORE_PATTERNS = [
    NAMES['mn_dir'],
]

class Repository(object):
    # Since click will try to instantiate this class for us with no args, we
    # put the __init__ code here instead
    def init(self, path):
        self.path = path
        self.index = Index(path)
        self.ignore = Ignores(DEFAULT_IGNORE_PATTERNS)

        # Other paths
        self.mn_dir_path = os.path.join(self.path, NAMES['mn_dir'])
        self.index_path = os.path.join(self.mn_dir_path, NAMES['index'])
        self.ignore_path = os.path.join(self.mn_dir_path, NAMES['ignore'])


    def create(self, path):
        repo_path = find_repo(path)

        if repo_path:
            output.error("Repository already exists: {}".format(repo_path))
            util.abort()

        self.init(path)

        os.mkdir(self.mn_dir_path)


    def open(self, path):
        if not os.path.exists(path):
            output.error("Repository path does not exist: {}".format(path))
            util.abort()
        elif not os.path.isdir(path):
            output.error("Repository path is not a directory: {}".format(path))
            util.abort()

        repo_path = find_repo(path)

        if not repo_path:
            output.error("Cannot find repository in '{}' or any parent dir".format(path))
            util.abort()

        self.init(repo_path)

        if os.path.isfile(self.index_path):
            self.index.read(self.index_path)

        if os.path.isfile(self.ignore_path):
            self.ignore.read(self.ignore_path)


def find_repo(start_path):
    if start_path == '/':
        return None

    mn_dir_path = os.path.join(start_path, NAMES['mn_dir'])

    if os.path.isdir(mn_dir_path):
        return start_path

    parent = os.path.split(start_path)[0]

    return find_repo(parent)



import click

GOOD_COLOR = 'green'
WARN_COLOR = 'yellow'
BAD_COLOR = 'red'
IGNORED_COLOR = 'blue'

def set_output_color(color):
    # Print nothing except the ANSI escape sequence
    click.secho('', nl=False, fg=color, reset=False)


def clear_output_color():
    # Print nothing except the reset escape sequence
    click.secho('', nl=False, reset=True)


def output(message, color=None, items=[]):
    if color:
        set_output_color(color)

    click.echo(message)

    for item in sorted(items):
        click.echo("  " + item)

    clear_output_color()

def info(message, items=[]):
    output(message, None, items)

def success(message, items=[]):
    output("SUCCESS: " + message, GOOD_COLOR, items)

def warning(message, items=[]):
    output("WARNING: " + message, WARN_COLOR, items)


def error(message, items=[]):
    output("ERROR: " + message, BAD_COLOR, items)

def print_filelists(new_files, changed_files, missing_files, ignored_files):
    if not any([new_files, changed_files, missing_files, ignored_files]):
        output("Index is up-to-date (no changes)", GOOD_COLOR)
        return

    if new_files:
        output("New files (not in index):", WARN_COLOR, new_files)

        # Print a blank space between sections
        if missing_files or changed_files or ignored_files:
            click.echo()

    if missing_files:
        output("Missing files:", WARN_COLOR, missing_files)

        # Print a blank space between sections
        if changed_files or ignored_files:
            click.echo()

    if changed_files:
        output("Changed files (hash differs from index):", BAD_COLOR, changed_files)

        # Print a blank space between sections
        if ignored_files:
            click.echo()

    if ignored_files:
        output("Ignored files and directories:", IGNORED_COLOR, ignored_files)



import os
from codecs import open
from setuptools import setup, find_packages

this_dir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_dir, 'DESCRIPTION.rst'), encoding='utf8') as f:
    long_desc = f.read()

setup(
    name='morenines',
    version='1.0.0',
    url='https://github.com/mcgid/morenines',
    license='MIT',

    description='A simple content change detector',
    long_description=long_desc,

    keywords=[
        'backup',
        'hashing',
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Topic :: System",
        "Topic :: System :: Archiving",
        "Topic :: System :: Archiving :: Mirroring",
        "Topic :: Utilities",
    ],

    install_requires=[
        'click',
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'mn = morenines.application:main'
        ]
    },
)



import os
from fnmatch import fnmatchcase
import click


class Ignores(object):
    def __init__(self, default_patterns=[]):
        self.patterns = default_patterns

    def read(self, path):
        with open(path, 'r') as stream:
            self.patterns.extend([line.strip() for line in stream])
    
    def match(self, path):
        filename = os.path.basename(path)

        if any(fnmatchcase(filename, pattern) for pattern in self.patterns):
            return True
        else:
            return False
